//----------------------------------------------------------------------------
//
//		FileEventsMonitor.cpp
//
//
//		Kernel / User Mode	:	User
//
//		
//		Responsible for 
//		- loading the filter driver
//		- opening a xml file
//		- wait for file events from the driver
//		- filter these and write the ones we want to the xml file.
//
//
// File History
//
//	Date			Person		Comment
//
//  7 Feb 2011		M Lewis		First Written
//
//----------------------------------------------------------------------------


#include "FileEventsMonitor.h"
#include "Strsafe.h"
#include "Tlhelp32.h"
#include "ConfigFile.h"

 

extern CConfigFile ConfigFile;



FileEventsMonitor::FileEventsMonitor( FILE* pLogFile, wchar_t* pXmlFilename )
{
	LogFile = pLogFile;

	XmlFilename.assign( pXmlFilename );

	bIsMonitorRunning		= false;
	bIsDriverInstalled		= false;
//	XmlFile					= NULL;
	FilterEventsByProcId	= 0;
	ProcCacheEntries		= 0;
	bCaptureAllEvents		= false;

	fileEvents = (BYTE*)malloc(FILE_EVENTS_BUFFER_SIZE);

	fileMonitorThread = new OOThreadHandler(this);


	hMonitorStoppedEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
	
	// open the xml file for the file events

	if ( ! XMLCreation.CreateAFileEventsDoc() )
	{
		fwprintf( LogFile, L"Error whilst creating the File Events XML DOM.\n" );	
	}

	// now load the filter driver into windows
	// and if it loads ok, then finish initialising
	// some data structures.

	if ( LoadFilterDriver() )
	{
		SetupDiskNameToDriveNameMap();

		LoadProcessFilterList();
	}
}







//------------------------------------------------------
// Class Destructor.
// do lots of cleaning up.


FileEventsMonitor::~FileEventsMonitor(void)
{
	StopMonitorThread();

	CloseHandle(hMonitorStoppedEvent);
	
	free(fileEvents);

	delete fileMonitorThread;

	if ( ! XMLCreation.SaveFileEventsDoc( XmlFilename.c_str() ) )
	{
		fwprintf( LogFile, L"Error whilst completing and writing the 'file events' xml file\n" );
	}

	if( bIsDriverInstalled )
	{
		bIsDriverInstalled = false;

		CloseHandle(communicationPort);

		FilterUnload( FILE_EVENT_FILTER_DRIVER_NAME );
	}
}



//---------------------------------------------------------
//
// Initiates the File Monitoring thread.
//


void FileEventsMonitor::StartMonitorThread()
{
	// dont bother to start the monitoring thread
	// if the driver didnt load. It will have nothing
	// to do.

	if ( bIsDriverInstalled == false ) return;
		
	// if the thread is already running, then do nothing.

	if( bIsMonitorRunning == true ) return;

	// start the event monitoring thread.
		
	fileMonitorThread->start();
}


//---------------------------------------------------------
//
// Sets a flag varible to signal to the file monitoring
// thread that it needs to shut down at the next available
// oportunity.
// we then wait for the thread to signal that it has indeed 
// stopped.
//


void FileEventsMonitor::StopMonitorThread()
{	
	// if the thread isnt running, there is nothing to do.

	if ( bIsMonitorRunning == false ) return;
	
	// set flag to indicate that thread should fall
	// out of its event processing loop.

	bIsMonitorRunning = false;
	
	// wait for the thread to clean up and close.

	WaitForSingleObject(hMonitorStoppedEvent, 5000);
	
}


//---------------------------------------------------------
//
// Convert the file event code into the text string so
// that it can be displayed or logged to a file.


bool FileEventsMonitor::GetEventName(PFILE_EVENT pFileEvent, wstring *fileEventName)
{
	bool found = true;
	*fileEventName = L"UNKNOWN";


	switch( pFileEvent->majorFileEventType ) 
	{
		case IRP_MJ_CREATE:
		{
			switch( pFileEvent->information )
			{
				case FILE_CREATED:
				case FILE_OVERWRITTEN:
				case FILE_SUPERSEDED:	*fileEventName = L"Create";	break;

				case FILE_OPENED:		*fileEventName = L"Open";	break;
		
				default:	found = false;
			}	
	
			wstring eventPath = pFileEvent->filePath;

			if(eventPath.find(L":", 0) != wstring::npos)  
			{
				found = false;
			}
		}
		break;

		case IRP_MJ_READ: 
			*fileEventName = L"Read";
			break;

		case IRP_MJ_WRITE: 
			*fileEventName = L"Write";
			break;

		case IRP_MJ_DELETE:
			*fileEventName = L"Delete";
			break;

		case IRP_MJ_RENAME:
			*fileEventName = L"Rename";
			break;

		default:
			found = false;
	}
	
	return found;
}



//-------------------------------------------------------------------
// function to convert
// Disk device names to Windows Disk Names
//
// eg.
//
// "\Device\HarddiskVolume1\aaaa"  --> "C:\aaaaa"
//


wstring FileEventsMonitor::ConvertRawFilePathameToDriveAndPathname(wstring fileObjectName)
{
	stdext::hash_map<wstring, wstring>::iterator it;

	for(it = dosNameMap.begin(); it != dosNameMap.end(); it++)
	{
		size_t position = fileObjectName.rfind(it->first,0);
		
		if(position != wstring::npos)
		{
			fileObjectName.replace(position, it->first.length(), it->second, 0, it->second.length());
			break;
		}
	}

	return fileObjectName;
}



//---------------------------------------------------------
// tests to see if a path refers to a directory.
//

bool FileEventsMonitor::isDirectory( wstring filePath )
{
	DWORD code = GetFileAttributes(filePath.c_str());

	if ( code == INVALID_FILE_ATTRIBUTES ) return false;

	if ( ( code & FILE_ATTRIBUTE_DIRECTORY ) == 0 ) return false;
	
	return true;
}





//-------------------------------------------------------------
//
// sets up a data structure that lists the names of processes
// that we wish to capture file events from.
//

void FileEventsMonitor::LoadProcessFilterList()
{
	ProcessList.clear();

	if ( ConfigFile.GetHashStringValuesFor( L"FileFilterProcessNames", ProcessList ) == false )
	{
		fwprintf( LogFile, L"Unable to get the file process names from the ini file\n" );
		wprintf( L"Unable to get the file process names from the ini file\n" );
	}

}


//-------------------------------------------------------------
//
// This function is the part of this class that is run as 
// a separate thread.
//
// It polls the File Filter Driver and recieves a buffer of 
// file events. These are parsed and the results written
// to an xml file.
// This continues until a flag variable is set indicating that
// the thread should shut down.
//



void FileEventsMonitor::run()
{
	// Get the full path to the working directory of this app.
	// we wish to exclude all file operations for this dir or
	// any subdirs.
	
	wchar_t AppDirectory[1024];	
	GetFullPathName(L".", 1024, AppDirectory, NULL);
	ThisAppWorkngDirectory = AppDirectory;

	transform( ThisAppWorkngDirectory.begin(), ThisAppWorkngDirectory.end(), ThisAppWorkngDirectory.begin(), tolower); 

//	wprintf(L"\napp dir = >%s< \n\n ",  AppDirectory );

	//

	HRESULT hResult;
	DWORD bytesReturned = 0;
	bIsMonitorRunning = true;

	UINT FilterParentProcess = 0;
	wchar_t* pFilterProcName = L" ";
	wchar_t  FileLongPath[1000];
	wchar_t  FileShortPath[1000];

	UCHAR	 LastEvent				= 0;
	wstring	 LastfileEventPath		= L"";
	DWORD	 LastProcessId			= 0;


	// enter the main processing loop of this thread.

	while( bIsMonitorRunning )
	{
		FILEMONITOR_MESSAGE command;
		command.Command = GetFileEvents;

		ZeroMemory(fileEvents, FILE_EVENTS_BUFFER_SIZE);

		hResult = FilterSendMessage( communicationPort,
                                     &command,
                                     sizeof( FILEMONITOR_COMMAND ),
                                     fileEvents,
                                     FILE_EVENTS_BUFFER_SIZE,
                                     &bytesReturned );

		if( hResult == S_OK && bytesReturned >= sizeof(FILE_EVENT))
		{
			UINT offset = 0;

			while(offset < bytesReturned)
			{
				FILE_EVENT* e = (FILE_EVENT*)(fileEvents + offset);

				if ( e->dataLost != 0 )
				{
					XMLCreation.WriteFileDataLostError();
					fwprintf( LogFile, L"\n!! File Filter Buffer was not read quickly enough. Data was Lost !!\n");
					wprintf( L"\n!! File Filter Buffer was not read quickly enough. Data was Lost !!\n");
				}
				else
				{

					wstring fileEventName;
					wstring fileEventPath;
					wstring file2EventPath = L"";

					UINT    ParentProcess = 0;
					UINT    GrandParentProcess = 0;
					UINT    GGrandParentProcess = 0;

					wchar_t* pProcName;
					wchar_t* pParentProcName;
					wchar_t* pGrandParentProcName;

					//----------------------------------------------
					// remove duplicate file WRITE messages.

					fileEventPath = e->filePath;

					if ( e->majorFileEventType == IRP_MJ_WRITE &&
						 LastEvent ==  e->majorFileEventType &&
						 LastfileEventPath == fileEventPath &&
						 LastProcessId == e->processId
					   )
					{
//						wprintf(L"Dupe %s\n",  fileEventPath );						
					}
					else
					{
						LastEvent			= e->majorFileEventType;
						LastfileEventPath	= fileEventPath;
						LastProcessId		= e->processId;

						//----------------------------------------------

						if( GetEventName(e, &fileEventName))
						{					
							fileEventPath = ConvertRawFilePathameToDriveAndPathname(fileEventPath);

							//----------------------------------------------

							if (  e->majorFileEventType == IRP_MJ_RENAME && e->file2PathLength > 0 )
							{
								void* p2ndFilename = ((char*) e->filePath ) + e->filePathLength;
								file2EventPath.assign( ((wchar_t*) p2ndFilename ) );

								// now strip off the "\??\" at the start
								// if it has been supplied by windows. 

								int pos = file2EventPath.find( L"\\??\\" );

								if ( pos == 0 )
								{
									file2EventPath = file2EventPath.substr( 4 );
								}

							}

							//----------------------------------------------

							wstring FirstNcharsOfPAth = fileEventPath.substr( 0, ThisAppWorkngDirectory.length() );

							transform( FirstNcharsOfPAth.begin(), FirstNcharsOfPAth.end(), FirstNcharsOfPAth.begin(), tolower); 

							if(  FilterEventsByProcId != 0 && 
								 fileEventPath != L"UNKNOWN" &&
								 ThisAppWorkngDirectory.compare( FirstNcharsOfPAth ) != 0 )  
							{
								ParentProcess = 0;
								GrandParentProcess = 0;
								GGrandParentProcess = 0;

								pProcName = L"";
								pParentProcName = L"";
								pGrandParentProcName = L"";

								FilterParentProcess = 1;
								pFilterProcName = L" ";

								CachedGetProcessNameAndParentId( e->processId, &ParentProcess, &pProcName );

								CachedGetProcessNameAndParentId( ParentProcess, &GrandParentProcess, &pParentProcName );

								CachedGetProcessNameAndParentId( GrandParentProcess, &GGrandParentProcess, &pGrandParentProcName );

								CachedGetProcessNameAndParentId( FilterEventsByProcId, &FilterParentProcess, &pFilterProcName );


								// have we found our process being monitored.. or a child of our monitored process,
								// or is it the msiexec process.. 
								// or is it a process with the same name as our monitored proc

								if (	bCaptureAllEvents ||
										FilterEventsByProcId == e->processId || 
										FilterEventsByProcId == ParentProcess || 
										FilterEventsByProcId == GrandParentProcess ||
										FilterEventsByProcId == GGrandParentProcess ||
										wcscmp( pFilterProcName, pProcName ) == 0 ||
										ProcessList.find( pProcName ) != ProcessList.end() ||
										ProcessList.find( pParentProcName ) != ProcessList.end() ||		// v important. need child processes of things such as MSIEXEC service
										ProcessList.find( pGrandParentProcName ) != ProcessList.end()	// v important. need child processes of things such as MSIEXEC service
									)  
								{

									if ( e->majorFileEventType == IRP_MJ_DELETE || 
										 ( e->majorFileEventType == IRP_MJ_CREATE && ( e->information == FILE_OVERWRITTEN || e->information == FILE_SUPERSEDED ) && ((e->status & 1) != 0) && e->flags !=0 ) ||  
										 ( e->majorFileEventType == IRP_MJ_CREATE && e->information == FILE_CREATED && e->flags !=0  ) ||  									   
										 e->majorFileEventType == IRP_MJ_WRITE ||
										   e->majorFileEventType == IRP_MJ_RENAME  ) 
									{

										wstring ShortFileName = fileEventPath;

										// If the file path is o a short filename, then try to convert it to a long filename

										if ( GetLongPathNameW( fileEventPath.c_str(), FileLongPath, 1000 ) != 0 )
										{
											fileEventPath = FileLongPath;
										}

										if ( GetShortPathNameW( ShortFileName.c_str(), FileShortPath, 1000 ) != 0 )
										{
											ShortFileName = FileShortPath;
										}

										// repeat for the 2nd filepath if it is a rename operation.

										if ( e->majorFileEventType == IRP_MJ_RENAME )
										{
											if ( GetLongPathNameW( file2EventPath.c_str(), FileLongPath, 1000 ) != 0 )
											{
												file2EventPath = FileLongPath;
											}
											else
											{
	//											fwprintf( LogFile, L"Long name 2 lookup for %s failed. err = %i\n", file2EventPath.c_str(), GetLastError() );
											}
										}

										// check for duplicate entries for create or writes..
										//
										// if it is not in the list carry on, else then ignore this event. 
										// we dont want duplicates in the xml log.
									
										if ( modifiedFiles.find( fileEventPath ) == modifiedFiles.end() 
											|| e->majorFileEventType == IRP_MJ_DELETE 
											|| e->majorFileEventType == IRP_MJ_RENAME )
										{
											// code to build a list of files modified whilst running this filter.

											if(!isDirectory(fileEventPath))
											{
												if( e->majorFileEventType == IRP_MJ_CREATE || 
													e->majorFileEventType == IRP_MJ_WRITE )
												{	
													modifiedFiles.insert(fileEventPath);
												} 
												else 
												if(e->majorFileEventType == IRP_MJ_DELETE) 
												{
													modifiedFiles.erase(fileEventPath);

													if ( ShortFileName != fileEventPath ) 
													{
														modifiedFiles.erase(ShortFileName);
													}
												}	
												else
												if(e->majorFileEventType == IRP_MJ_RENAME )
												{
													modifiedFiles.erase(fileEventPath);

													if ( ShortFileName != fileEventPath ) 
													{
														modifiedFiles.erase(ShortFileName);
													}

													modifiedFiles.insert(file2EventPath);
												}

											}

							

  											wchar_t procid[100], datetime[100];

											StringCbPrintf( procid, sizeof(TCHAR)*100, L"%d", e->processId );
											StringCbPrintf( datetime, sizeof(TCHAR)*100, L"%lld", e->time.QuadPart );

											XMLCreation.WriteFileEvent( procid, pProcName, datetime, fileEventName.c_str(), fileEventPath.c_str(), file2EventPath.c_str() );

					//							fwprintf( File, L"Information %d \t Minor op : %d \t status : %x \t flags : %x \n", e->information, e->minorFileEventType,e->status, e->flags );

										}	// end of check for duplicate entries
									}		// end of check for event type of interest
								}			// end of check for an event of interest because process is of interest
							}				// end of check for valid event and not in app directory
						}					// end of check for geteventname
					}						// end of check for duplicate WRITE events
				}							// end of check for loss of data.
				
				offset += sizeof(FILE_EVENT) + e->filePathLength + e->file2PathLength;
			} 
		}
		
		Sleep(FILE_EVENT_WAIT_TIME);
	}

	// signal that the monitoring thread is shutting down.

	SetEvent(hMonitorStoppedEvent);
}




//-------------------------------------------------------------
//
// communicates with the File Filter Driver.
//
// Tells it that we want to keep copies of files which get deleted
// during the app installation.
// We send the directory where we want the files to be stored.
//



void FileEventsMonitor::SetDeletedFilesDir( BOOLEAN OnOff, wchar_t* DeletedFilesDir )
{
	if ( bIsDriverInstalled == false ) return;

	FILEMONITOR_MESSAGE command;
	FILEMONITOR_SETUP	setupFileMonitor;
	HRESULT				hResult;
	DWORD				bytesReturned;

	setupFileMonitor.bCollectDeletedFiles = OnOff;
	
	GetFullPathName( DeletedFilesDir, 1024, setupFileMonitor.wszLogDirectory, NULL);
	
	CreateDirectory(setupFileMonitor.wszLogDirectory, NULL );
	
	setupFileMonitor.nLogDirectorySize = (UINT)wcslen(setupFileMonitor.wszLogDirectory)*sizeof(wchar_t);
	
	command.Command = SetupMonitor;	

	hResult = FilterSendMessage( communicationPort,
							 &command,
							 sizeof( FILEMONITOR_COMMAND ),
							 &setupFileMonitor,
							 sizeof(FILEMONITOR_SETUP),
							 &bytesReturned );
}


//------------------------------------------------------
//
// Get the Process Name and the parent proc id of the process
// specified by paramter 'pid'.
//
// This version caches results so that it will be much faster
// than using system api calls each time.
//




BOOL FileEventsMonitor::CachedGetProcessNameAndParentId(int pid, UINT* pParent, wchar_t** pProcName )
{
	UINT prnt; 
	wchar_t ProcessName[100];

	UINT i;

	for( i = 0; i < ProcCacheEntries; i++ )
	{
		if ( pid == ProcId[ i ] )
		{
			*pProcName = ProcName[ i ];
			*pParent   = ProcParentId[ i ];

		    return TRUE;
		}
	}

		
	if ( GetProcessNameAndParentId( pid, &prnt, ProcessName ) == false )
	{
		return false;
	}

	if ( ProcCacheEntries + 1 < PROCESS_PARENT_CACHE_SIZE )
	{
		ProcId[ ProcCacheEntries ] = pid;
		ProcParentId[ ProcCacheEntries ] = prnt;	
		wcscpy_s( ProcName[ ProcCacheEntries ], MAX_PROC_NAME_SIZE, ProcessName );

		*pProcName = (wchar_t*) ProcName[ ProcCacheEntries ];
		*pParent = prnt;

		ProcCacheEntries++;

//		if ( prnt != 0 ) printf("proc %d is parent of %d\n", prnt, pid );

		return TRUE;
	}
	else
	{
		fwprintf( LogFile, L"Proc / Proc Parent Cache is full !\n");
		return FALSE;
	}

	return FALSE;
}



//------------------------------------------------------
//
// Get the Process Name and the parent proc id of the process
// specified by paramter 'pid'.
//
// This calls the windows api to get the info. 
// This can be a performance bottleneck when called very frequently,
// so a cached version is provided instead. ( That routine calls this 
// one to get entries for its cache )
//


BOOL FileEventsMonitor::GetProcessNameAndParentId(int pid, UINT* pParent, wchar_t* pProcName)
{
//	printf("get parent called for %d \n", pid );

	HANDLE hSnapShot = CreateToolhelp32Snapshot( TH32CS_SNAPPROCESS, 0);

	if ( hSnapShot == INVALID_HANDLE_VALUE )
	{
		fwprintf( LogFile, L"Error getting processes snapshot" );
		return FALSE;
	}

	PROCESSENTRY32* procentry = new PROCESSENTRY32();
	
	procentry->dwSize = sizeof( PROCESSENTRY32 );


	BOOL bContinue = Process32First(hSnapShot, procentry);

	while (bContinue)
	{
		if (pid == procentry->th32ProcessID)
		{
//			wprintf( L"%d is parent of %d : %s \n", procentry->th32ParentProcessID, pid, procentry->szExeFile );
			
			wcscpy_s( pProcName, MAX_PROC_NAME_SIZE, procentry->szExeFile );

			*pParent = procentry->th32ParentProcessID;

			return TRUE;
		}

		bContinue = Process32Next(hSnapShot, procentry);
	}

	return FALSE;
}


//------------------------------------------------------------------------


void FileEventsMonitor::DumpProcessRelationshipsCache()
{
	fwprintf( LogFile, L"\n\nProcess Relationships Dump\n\n" );
	
	for( UINT i = 0; i < ProcCacheEntries; i++ )
	{
		UINT Parent;
		wchar_t* ParentName;
		ParentName = L" ";

		CachedGetProcessNameAndParentId( ProcParentId[ i ], &Parent, &ParentName );

		fwprintf( LogFile, L"Process %i \t %25s \t has parent %i \t %s\n", ProcId[ i ], ProcName[ i ], ProcParentId[ i ], ParentName );
	}

	fwprintf( LogFile, L"\n\n" );
}


//------------------------------------------------------------------------


bool FileEventsMonitor::LoadFilterDriver()
{
	FilterUnload( FILE_EVENT_FILTER_DRIVER_NAME );

	HRESULT hResult = FilterLoad( FILE_EVENT_FILTER_DRIVER_NAME );

	for( int count = 1; ( count < 5) && IS_ERROR( hResult ); count++ )
	{
		fwprintf( LogFile, L"File filter driver did not load ( error: %x ). will retry in a few seconds\n", hResult );

		Sleep(2000);

		hResult = FilterLoad( FILE_EVENT_FILTER_DRIVER_NAME );
	}
	

	if (!IS_ERROR( hResult )) 
	{
		
		hResult = FilterConnectCommunicationPort( CAPTURE_FILEMON_PORT_NAME,
												0,
												NULL,
												0,
												NULL,
												&communicationPort );
		if (IS_ERROR( hResult )) 
		{
			fwprintf( LogFile, L"Could not connect to file filter: 0x%08x\n", hResult );
		} 
		else 
		{
			fwprintf( LogFile, L"Loaded kernal file filter driver\n");
			bIsDriverInstalled = true;
		}		
	}

	return bIsDriverInstalled;
}




//-------------------------------------------------------
//
// copies monitored files that have been modified by the
// app installer. These files are copied to a file cache
// directory tree in the output directory.

// error codes.
//
// 2	= The system cannot find the file specified.
// 3	= The system cannot find the path specified.
// 4	= The system cannot open the file.


void FileEventsMonitor::CopyModifiedFilesToOutputDir( wchar_t WrittenFilesDestDir[] )
{
	wchar_t WrittenFilesDestPath[1024];
	wchar_t ErrMsg[1000];

	GetFullPathName( WrittenFilesDestDir, 1024, WrittenFilesDestPath, NULL);

	stdext::hash_set<wstring>::iterator it;
	
	wstring WrittenFilesDir = WrittenFilesDestPath;
	
	for(it = modifiedFiles.begin(); it != modifiedFiles.end(); it++)
	{		
		wstring SrcFilePath	= *it;

		int FirstSlashPos	= SrcFilePath.find_first_of(L"\\");
		int LastSlashPos	= SrcFilePath.find_last_of(L"\\");

		wstring SrcDrive	= SrcFilePath.substr( 0, SrcFilePath.find_first_of(L":") );
		wstring SrcFileName	= SrcFilePath.substr( LastSlashPos+1 );
		wstring SrcDirs     = SrcFilePath.substr( FirstSlashPos+1, LastSlashPos-FirstSlashPos-1 );

		// create a path for where we want to copy the file.

		wstring DestDir = WrittenFilesDir + L"\\" + SrcDrive + L"\\" + SrcDirs;

		SHCreateDirectoryEx(NULL, DestDir.c_str(), NULL );

		// add the filename on the end of the dir path, to create a complete
		// path for the file.

		wstring DestFilePathAndName = DestDir + L"\\" + SrcFileName;
	
		// copy the file from its source to out output directory.

		if( CopyFile( SrcFilePath.c_str(), DestFilePathAndName.c_str(), FALSE ) )
		{
			fwprintf( LogFile, L"copied %s\n",  SrcFilePath.c_str() );
		}
		else
		{
			int ErrorId = GetLastError();
			ErrMsg[0] = 0;

			FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM, 0, ErrorId, 0, ErrMsg, 1000, 0 );
	
			fwprintf( LogFile, L"  >>>> unable to copy file %s : error code = 0x%08x : %s \n", SrcFilePath.c_str(), ErrorId, ErrMsg );	
		} 

	}

	modifiedFiles.clear();
}



//------------------------------------------------------
//
// Create a structure which holds mappings from 
// Disk device names to Windows Disk Names
//
// eg.
//
// "\Device\HarddiskVolume1\"  --> "C:\"
//


void FileEventsMonitor::SetupDiskNameToDriveNameMap()
{
	UCHAR	buffer[1024];
	HANDLE	SearchHandle = INVALID_HANDLE_VALUE;
	ULONG	volumeBytesReturned;
	WCHAR	driveLetter[10];


	PFILTER_VOLUME_BASIC_INFORMATION pVolumeBuff = (PFILTER_VOLUME_BASIC_INFORMATION)buffer;

	HRESULT hResult = FilterVolumeFindFirst( FilterVolumeBasicInformation,
											 pVolumeBuff,
											 sizeof(buffer)-sizeof(WCHAR),   //save space to null terminate name
											 &volumeBytesReturned,
											 &SearchHandle );

	if ( SearchHandle == INVALID_HANDLE_VALUE ) return;

	while( SUCCEEDED( hResult ) )
	{
		// null terminate the string in the buffer.

		pVolumeBuff->FilterVolumeName[ pVolumeBuff->FilterVolumeNameLength/sizeof( WCHAR ) ] = UNICODE_NULL;
		
		// now go and try to see if there is an associated DOS drive name for this device.
		// if there is, then include it in the device to drive name map.

		if( SUCCEEDED( FilterGetDosName( pVolumeBuff->FilterVolumeName, driveLetter, sizeof(driveLetter)/sizeof(WCHAR) ) ))
		{
			dosNameMap.insert( DosPair( pVolumeBuff->FilterVolumeName, (wstring) driveLetter ) );
		}
		
		// now go and get the next device.

		hResult = FilterVolumeFindNext( SearchHandle,
										FilterVolumeBasicInformation,
										pVolumeBuff,
										sizeof(buffer)-sizeof(WCHAR),    //save space to null terminate name
										&volumeBytesReturned );
	}

	// if the Handle is still valid, then close it.

	if ( SearchHandle != INVALID_HANDLE_VALUE ) 
	{
		FilterVolumeFindClose( SearchHandle );
	}
}


