//----------------------------------------------------------------------------
//
//		RegEventsMonitor.cpp
//
//
//		Kernel / User Mode	:	User
//
//		
//		Responsible for 
//		- loading the filter driver
//		- opening a xml file
//		- wait for registry events from the driver
//		- filter these and write the ones we want to the xml file.
//
//
// File History
//
//	Date			Person		Comment
//
//  30 Jan 2011		M Lewis		First Written
//
//----------------------------------------------------------------------------


#include "RegEventsMonitor.h"
#include "Tlhelp32.h"
#include <algorithm>
#include "Strsafe.h"
#include "ConfigFile.h"


extern CConfigFile ConfigFile;

//------------------------------------------------------
// Class Constructor
//
// Opens up an XML log file for results.
//
// sets up data structures ready for when the monitor thread is 
// started.
//
// Also loads in the Filter driver to monitor the registry
//


RegEventsMonitor::RegEventsMonitor( FILE* pLogFile, wchar_t* pXmlFilename )
{
	LogFile = pLogFile;

	XmlFilename.assign( pXmlFilename );

	FilterEventsByProcId	= 0;
	hService				= NULL;
	hDriverComms			= NULL;
	bRegDriverLoaded		= false;
	bThreadRunning			= false;
	numRegMappings			= 0;
	ProcCacheEntries		= 0;
	bCaptureAllEvents		= false;

	// create the buffer used to fetch data back from the Filter
	// Driver.

	RegEventsBuffer = (BYTE*)malloc( REGISTRY_EVENTS_BUFFER_SIZE );

	// create a helper class to allow us to run threads in
	// an OO way.

	registryMonitorThread = new OOThreadHandler(this);

	hThreadStoppedEvent = CreateEvent(NULL, FALSE, FALSE, NULL);

	// open an XML file for logging the registry events

	if ( ! XMLCreation.CreateARegEventsDoc() )
	{
		fwprintf( LogFile, L"Error whilst creating the Reg Events XML DOM.\n" );	
	}

	LoadFilterDriver();

	SetupPartialRegPathToHiveNameMap();

	LoadProcessFilterList();
}





//------------------------------------------------------
// Class Destructor.
// do lots of cleaning up.


RegEventsMonitor::~RegEventsMonitor(void)
{
	StopMonitorThread();

	if ( ! XMLCreation.SaveRegEventsDoc( XmlFilename.c_str() ) )
	{
		fwprintf( LogFile, L"Error whilst completing and writing the 'registry events' xml file\n" );
	}

	CloseHandle( hThreadStoppedEvent );

	free( RegEventsBuffer );

	delete registryMonitorThread;

	if( bRegDriverLoaded )
	{
		bRegDriverLoaded = false;
		if ( hDriverComms != NULL ) CloseHandle(hDriverComms);
	}

	UninstallRegistryKernelDriver();
}



//-------------------------------------------------------------
//
// sets up a data structure that lists the names of processes
// that we wish to capture Registry events from.
//

void RegEventsMonitor::LoadProcessFilterList()
{
	ProcessList.clear();

	if ( ConfigFile.GetHashStringValuesFor( L"RegistryFilterProcessNames", ProcessList ) == false )
	{
		fwprintf( LogFile, L"Unable to get the registry process names from the ini file\n" );
		wprintf( L"Unable to get the registry process names from the ini file\n" );
	}

}



//------------------------------------------------------
//
// starts the Monitor Thread


void RegEventsMonitor::StartMonitorThread()
{
	// dont bother to start the monitoring thread
	// if the driver didnt load. It will have nothing
	// to do.

	if ( bRegDriverLoaded == false ) return;

	// if the thread is already running, then do nothing.

	if ( bThreadRunning == true ) return;

	// start the event monitoring thread.

	registryMonitorThread->start();
}



//------------------------------------------------------
//
// Stops the Monitor Thread.
//


void RegEventsMonitor::StopMonitorThread()
{
	// if the thread isnt running, there is nothing to do.

	if ( bThreadRunning == false ) return; 

	// set flag to indicate that thread should fall
	// out of its event processing loop.

	bThreadRunning = false;

	// wait for the thread to clean up and close.

	WaitForSingleObject(hThreadStoppedEvent, 5000);	
}



//------------------------------------------------------
//
// This function gets the raw path in the registry for
// the current user.
//
// eg. "\REGISTRY\USER\S-1-5-21-1292428093-1390067357-1177238915-1004\
//



BOOL RegEventsMonitor::GetRegistryCurrentUserKey( wstring* psKey )
{
	HKEY hUserKey;

	wchar_t szUserKeyBuffer[500];

	PUBLIC_OBJECT_TYPE_INFORMATION* pUserKeyBuffer = (PUBLIC_OBJECT_TYPE_INFORMATION*) &szUserKeyBuffer;
	

	if (RegOpenCurrentUser(KEY_READ , &hUserKey) == ERROR_SUCCESS )
	{
		DWORD RequiredLength;

		pNtQueryObject NtQueryObject = (pNtQueryObject)GetProcAddress(GetModuleHandle(L"ntdll.dll"), (LPCSTR)"NtQueryObject");
		
		NTSTATUS status = NtQueryObject( hUserKey, 1, NULL, 0, &RequiredLength );

		if(status == STATUS_INFO_LENGTH_MISMATCH && RequiredLength < 256 )
		{
			if( status != NtQueryObject( hUserKey, 1, &szUserKeyBuffer, RequiredLength, &RequiredLength ) )
			{
				psKey->assign( pUserKeyBuffer->TypeName.Buffer, (pUserKeyBuffer->TypeName.Length / 2)  );

				RegCloseKey( hUserKey );

				return TRUE;
			}
		}

		RegCloseKey( hUserKey );
	}

	return FALSE;
}


//------------------------------------------------------
//
// Initialises a lookup list so that Partial Reg paths
// can be mapped to more understandable HIV names
//
// see ConvertPartialRegPathToHiveName()
//


void RegEventsMonitor::SetupPartialRegPathToHiveNameMap()
{
	wstring CUKey;

	if ( GetRegistryCurrentUserKey( &CUKey ) )
	{
		RegPartialPath[ numRegMappings	 ] = CUKey + L"_CLASSES";
		RegHiveName	  [ numRegMappings++ ] = L"HKCR";

		RegPartialPath[ numRegMappings   ] = CUKey;
		RegHiveName	  [ numRegMappings++ ] = L"HKCU";

		RegPartialPath[ numRegMappings   ] = L"\\REGISTRY\\MACHINE";
		RegHiveName	  [ numRegMappings++ ] = L"HKLM";

		RegPartialPath[ numRegMappings   ] = L"\\REGISTRY\\USER";
		RegHiveName	  [ numRegMappings++ ] = L"HKU";
	}
}



//------------------------------------------------------
//
// The converts raw registry path names to better known 
// Hive names
//
// eg. 
//		"\REGISTRY\MACHINE" is substituted with "HKLM"
// or
//      "\REGISTRY\USER\S-1-5-21-1292428093-1390067357-1177238915-1004\"
//      is substituted with 
//		"HKCU\"
//



wstring RegEventsMonitor::ConvertPartialRegPathToHiveName(wstring RegistryPath )
{	
	// convert the path to upper before comparing.
	
	wstring UpperRegPath = RegistryPath;

	transform
	(
		UpperRegPath.begin(), 
		UpperRegPath.end(),
		UpperRegPath.begin(),
		towupper
	);

	for( UINT i = 0; i < numRegMappings; i++)
	{
		size_t position = UpperRegPath.rfind( RegPartialPath[i], 0 );

		if( position != wstring::npos )
		{
			wstring NewRegPath = RegistryPath.replace( position, RegPartialPath[i].length(), RegHiveName[i], 0, RegHiveName[i].length() );
			return NewRegPath;
		}
	}

	return RegistryPath;
}



//------------------------------------------------------
//
// This function is executed as a worker thread.
//
// Its purpose is to communicate with the Registry Filter Driver
// and request a series of Registry Events. 
// These events are filtered and the ones that we are interested in
// are logged in an XML log file.
//



void RegEventsMonitor::run()
{
	bThreadRunning			 = true;

	UINT ParentProcess		 = 0;
	UINT GrandParentProcess = 0;
	UINT GGrandParentProcess = 0;
	UINT FilterParentProcess = 0;

	wchar_t* pProcName;
	wchar_t* pParentProcName;
	wchar_t* pGrandParentProcName;
	wchar_t* pFilterProcName;

	DWORD SizeOfBufferReturnedByFilter; 

	while( bThreadRunning )
	{
		ZeroMemory( RegEventsBuffer, REGISTRY_EVENTS_BUFFER_SIZE);

		SizeOfBufferReturnedByFilter = 0;

		// send a request to the kernal driver asking it to
		// fill the buffer up with Registry Events

		DeviceIoControl
		(
			hDriverComms,
			IOCTL_GET_REGISTRY_EVENTS, 
			0, 
			0, 
			RegEventsBuffer, 
			REGISTRY_EVENTS_BUFFER_SIZE, 
			&SizeOfBufferReturnedByFilter, 
			NULL
		);

		/* Go through all the registry events received. Events are variable sized
		   so the starts of them are calculated by adding the lengths of the various
		   data stored in it */
		
		if( SizeOfBufferReturnedByFilter >= sizeof(REG_EVENT_RECORD) )
		{

			UINT BufferOffset = 0;
			
			while( BufferOffset < SizeOfBufferReturnedByFilter )
			{
				// get a pointer to the record of interest in the buffer.

				REG_EVENT_RECORD* pEvent = (REG_EVENT_RECORD*)( RegEventsBuffer + BufferOffset );
				
				if ( pEvent->dataLost != 0 )
				{
					XMLCreation.WriteRegDataLostError();
					fwprintf( LogFile, L"\n!! Reg Filter Buffer was not read quickly enough. Data was Lost !!\n");
					wprintf( L"\n!! Reg Filter Buffer was not read quickly enough. Data was Lost !!\n");
				}
				else
				{
					wstring registryEventName = GetRegistryEventTypeName( pEvent->eventType );
				
					/* Get a pointer to the registry path string */
				
					wchar_t* pRegistryPath = (wchar_t*) pEvent->registryData;

					wstring registryPath = L"";
					wstring ValueName = L"";



					ParentProcess = 0;
					GrandParentProcess = 0;
					GGrandParentProcess = 0;

					pProcName = L"";
					pParentProcName = L"";
					pGrandParentProcName = L"";

					FilterParentProcess = 1;
					pFilterProcName = L" ";

					CachedGetProcessNameAndParentId( pEvent->processId, &ParentProcess, &pProcName );

					CachedGetProcessNameAndParentId( ParentProcess, &GrandParentProcess, &pParentProcName );

					CachedGetProcessNameAndParentId( GrandParentProcess, &GGrandParentProcess, &pGrandParentProcName );

					CachedGetProcessNameAndParentId( FilterEventsByProcId, &FilterParentProcess, &pFilterProcName );

					if ( bCaptureAllEvents ||
							( FilterEventsByProcId != 0 && 
					
								( FilterEventsByProcId == pEvent->processId || 
								  FilterEventsByProcId == ParentProcess ||
								  FilterEventsByProcId == GrandParentProcess ||
								  FilterEventsByProcId == GGrandParentProcess ||
								  wcscmp( pFilterProcName, pProcName ) == 0 || 
								  ProcessList.find( pProcName ) != ProcessList.end() ||
								  ProcessList.find( pParentProcName ) != ProcessList.end() ||		// v important. need child processes of things such as MSIEXEC service
								  ProcessList.find( pGrandParentProcName ) != ProcessList.end()		// v important. need child processes of things such as MSIEXEC service
								) 
							)
					   )
					{

						if ( pEvent->eventType == RegNtPostCreateKey || 
							 pEvent->eventType == RegNtPreSetValueKey || 
							 pEvent->eventType == RegNtPreDeleteKey ||
							 pEvent->eventType == RegNtDeleteValueKey ) 
						{
	
								wstring DataType = L"";
								wstring RegData = L"";

								// if there is a name as well as a key path, separate them out.
								// else just use the key path.

								if ( pEvent->registryPathKeyLengthB > 0 )
								{
									wstring RawRegPath = pRegistryPath;

									ValueName = RawRegPath.substr( (pEvent->registryPathKeyLengthB / 2)+1 );

									RawRegPath = RawRegPath.substr( 0, (pEvent->registryPathKeyLengthB / 2) );

									registryPath = ConvertPartialRegPathToHiveName( RawRegPath );
								}
								else
								{
									registryPath = ConvertPartialRegPathToHiveName( pRegistryPath );
								}


								// If there is data being written to the reg, then get it out of the event message.


								if ( pEvent->eventType == RegNtPreSetValueKey )
								{
									// get a pointer to the registry data 

									BYTE* pRegistryData = (BYTE*) pEvent->registryData + pEvent->registryPathLengthB;

									ExtractRegData( pEvent->dataType, pRegistryData, pEvent->dataLengthB, &DataType, &RegData );
								}


								// new xml generation code..

								wchar_t procid[100], datetime[100];

								StringCbPrintf( procid, sizeof(TCHAR)*100, L"%d", pEvent->processId );
								StringCbPrintf( datetime, sizeof(TCHAR)*100, L"%lld", pEvent->time.QuadPart );


								XMLCreation.WriteRegEvent( procid, pProcName, datetime, registryEventName.c_str(), registryPath.c_str(), ValueName.c_str(), DataType.c_str(), RegData.c_str() );

						}
					}
				}

				// add the size of this event record to the buffer offset

				BufferOffset += sizeof(REG_EVENT_RECORD) + pEvent->registryPathLengthB + pEvent->dataLengthB;
			} 
		}

		Sleep( REGISTRY_MONITOR_THREAD_SLEEP_TIME );
	}

	// this thread has been requested to shut down.
	// signal that this has actually happended.

	SetEvent( hThreadStoppedEvent );
}






//------------------------------------------------------
//
// converts an integer registry event code to a string
// description.
//


wstring RegEventsMonitor::GetRegistryEventTypeName(int EventType)
{
	wstring sEventName = L"?";

	switch( EventType )
	{
		case RegNtPostCreateKey:		sEventName = L"CreateKey";		break;
		case RegNtPreDeleteKey:			sEventName = L"DeleteKey";		break;
		case RegNtDeleteValueKey:		sEventName = L"DeleteValueKey"; 	break;
		case RegNtPreSetValueKey:		sEventName = L"SetValueKey";		break;
		case RegNtKeyHandleClose:		sEventName = L"CloseKey";		break;
		
		default:
			break;
	}
	return sEventName;
}




//------------------------------------------------------
//
// Formats Some Registy data from binry form to text
// so it can be written to the log.
//


DWORD RegEventsMonitor::ExtractRegData( ULONG Type, BYTE* registryData, ULONG dataLengthB, wstring* DataType, wstring* RegData  )
{
	PULONGLONG pqword;
	PULONG pulong;
	PTSTR  pszz = 0;
	wstring DataString;
	wchar_t tmps[100];


	switch( Type ) 
	{
	case REG_NONE:
		*DataType = L"NONE";
		break;

	case REG_SZ:    
		*DataType = L"SZ";
		RegData->assign( (wchar_t*) registryData );
		break;

	case REG_EXPAND_SZ:
		*DataType = L"EXPAND_SZ";
		RegData->assign( (wchar_t*) registryData );
		break;
                            
	case REG_BINARY:           
		*DataType = L"BINARY";
		for( UINT i = 0; i < dataLengthB; i++ )
		{
			if ( i > 0 ) *RegData = *RegData + L" ";
			StringCbPrintf( tmps, sizeof(TCHAR)*100, L"%02x", registryData[i] );
			*RegData = *RegData + wstring( tmps );
		}
		break;

	case REG_DWORD:                   
//	case REG_DWORD_LITTLE_ENDIAN:		// same as DWORD
		*DataType = L"DWORD";
		pulong = (PULONG) registryData;
		StringCbPrintf( tmps, sizeof(TCHAR)*100, L"0x%x", *pulong );
		RegData->assign( tmps );
		break;

	case REG_DWORD_BIG_ENDIAN: 
		{
			*DataType = L"DWORD_BIG_ENDIAN";
			DWORD tmp;
	
			pulong = (PULONG) registryData;

			DWORD dw = *pulong;

			tmp =  (dw & 0x000000FF);
			tmp = ((dw & 0x0000FF00) >> 0x08) | (tmp << 0x08);
			tmp = ((dw & 0x00FF0000) >> 0x10) | (tmp << 0x08);
			tmp = ((dw & 0xFF000000) >> 0x18) | (tmp << 0x08);

			StringCbPrintf( tmps, sizeof(TCHAR)*100, L"0x%x", tmp );
			RegData->assign( tmps );
		}
		break;

	case REG_LINK:     
		*DataType = L"LINK";
		RegData->assign( (wchar_t*) registryData );
		break;

	case REG_MULTI_SZ:    
		*DataType = L"MULTI_SZ";

		pszz = (PTSTR) registryData;
		while (*pszz) 
		{
			if ( pszz != (PTSTR) registryData ) *RegData = *RegData + L"\t";

			wstring onestring;

			onestring.assign( (wchar_t*) pszz );

			*RegData = *RegData + onestring;

			pszz = pszz + _tcslen(pszz) + 1;
		}
		break;

	case REG_RESOURCE_LIST: 
		*DataType = L"RESOURCE_LIST";

		for( UINT i = 0; i < dataLengthB; i++ )
		{
			if ( i > 0 ) *RegData = *RegData + L" ";
			StringCbPrintf( tmps, sizeof(TCHAR)*100, L"%02x", registryData[i] );
			*RegData = *RegData + wstring( tmps );
		}
		break;

	case REG_FULL_RESOURCE_DESCRIPTOR:
		*DataType = L"FULL_RESOURCE_DESCRIPTOR";

		for( UINT i = 0; i < dataLengthB; i++ )
		{
			if ( i > 0 ) *RegData = *RegData + L" ";
			StringCbPrintf( tmps, sizeof(TCHAR)*100, L"%02x", registryData[i] );
			*RegData = *RegData + wstring( tmps );
		}
		break;

	case REG_RESOURCE_REQUIREMENTS_LIST: 
		*DataType = L"FULL_RESOURCE_REQUIREMENTS_LIST";

		for( UINT i = 0; i < dataLengthB; i++ )
		{
			if ( i > 0 ) *RegData = *RegData + L" ";
			StringCbPrintf( tmps, sizeof(TCHAR)*100, L"%02x", registryData[i] );
			*RegData = *RegData + wstring( tmps );
		}
		break;

	case REG_QWORD:
//	case REG_QWORD_LITTLE_ENDIAN:				// same as REG_QWORD
		*DataType = L"QWORD";
		pqword = (PULONGLONG) registryData;
		StringCbPrintf( tmps, sizeof(TCHAR)*100, L"%I64x", *pqword );
		RegData->assign( tmps );
		break;

	default:
		*DataType = L"UNKNOWN";
		break;
	};

	return 0;
}




//------------------------------------------------------
//
// Get the Process Name and the parent proc id of the process
// specified by paramter 'pid'.
//
// This version caches results so that it will be much faster
// than using system api calls each time.
//

BOOL RegEventsMonitor::CachedGetProcessNameAndParentId(int pid, UINT* pParent, wchar_t** pProcName )
{
	UINT prnt; 
	wchar_t ProcessName[ MAX_PROC_NAME_SIZE ];

	int i;

	for( i = 0; i < ProcCacheEntries; i++ )
	{
		if ( pid == ProcCacheProcId[ i ] )
		{
			*pProcName = ProcCacheProcName[ i ];
			*pParent   = ProcCacheProcParentId[ i ];

		    return TRUE;
		}
	}

		
	if ( GetProcessNameAndParentId( pid, &prnt, ProcessName ) == false ) return false;

	if ( ProcCacheEntries + 1 < PROCESS_PARENT_CACHE_SIZE )
	{
		ProcCacheProcId[ ProcCacheEntries ] = pid;
		ProcCacheProcParentId[ ProcCacheEntries ] = prnt;	

		wcscpy_s( ProcCacheProcName[ ProcCacheEntries ], MAX_PROC_NAME_SIZE, ProcessName );

		*pProcName = (wchar_t*) ProcCacheProcName[ ProcCacheEntries ];
		*pParent = prnt;

		ProcCacheEntries++;

//		if ( prnt != 0 ) printf("proc %d is parent of %d\n", prnt, pid );

		return TRUE;
	}
	else
	{
		fwprintf( LogFile, L"Proc / Proc Parent Cache is full !\n");
		return FALSE;
	}

	return FALSE;
}



//------------------------------------------------------
//
// Get the Process Name and the parent proc id of the process
// specified by paramter 'pid'.
//
// This calls the windows api to get the info. 
// This can be a performance bottleneck when called very frequently,
// so a cached version is provided instead. ( That routine calls this 
// one to get entries for its cache )
//

BOOL RegEventsMonitor::GetProcessNameAndParentId(int pid, UINT* pParent, wchar_t* pProcName)
{

	HANDLE hSnapShot = CreateToolhelp32Snapshot( TH32CS_SNAPPROCESS, 0);

	if ( hSnapShot == INVALID_HANDLE_VALUE )
	{
		fwprintf( LogFile, L"Error getting processes snapshot" );
		return FALSE;
	}

	PROCESSENTRY32* procentry = new PROCESSENTRY32();
	
	procentry->dwSize = sizeof( PROCESSENTRY32 );


	BOOL bContinue = Process32First(hSnapShot, procentry);

	while (bContinue)
	{
		if (pid == procentry->th32ProcessID)
		{
//			wprintf( L"%d is parent of %d : %s \n", procentry->th32ParentProcessID, pid, procentry->szExeFile );
			
			wcscpy_s( pProcName, MAX_PROC_NAME_SIZE, procentry->szExeFile );

			*pParent = procentry->th32ParentProcessID;

			return TRUE;
		}

		bContinue = Process32Next(hSnapShot, procentry);
	}

	return FALSE;
}



//------------------------------------------------------------------
//
// Install the filter driver as a service, and start it.
//
// Then open a 'file' which is in the Win32 Device Namespace of
// the Registry Monitor Filter Driver.
//
// \\.\ = are the chars that define that a file name is a mapping
// to a device namespace.
//
// The 'file' created is the communication channel between this
// User mode code and the Filter Driver.
//



bool RegEventsMonitor::LoadFilterDriver()
{
	bRegDriverLoaded = false;

	// install the driver as a service

	if ( InstallRegistryKernelDriverAsService() )
	{
		if ( StartRegistryKernelDriverService() )
		{
			bRegDriverLoaded = true;
		}
		else
		{
			UninstallRegistryKernelDriver();
		}
	}

	
	// if the driver loaded ok, then try to open up the communication buffer

	if( bRegDriverLoaded )
	{	
		hDriverComms = CreateFile
					   (
						REG_EVENT_FILTER_DRIVER_NAMESPACE,
						GENERIC_READ | GENERIC_WRITE, 
						FILE_SHARE_READ | FILE_SHARE_WRITE,
						0,										// Default security
						OPEN_EXISTING,
						FILE_FLAG_OVERLAPPED,					// Perform asynchronous I/O
						0
					   );										// No template
		

		if( hDriverComms == INVALID_HANDLE_VALUE ) 
		{
			hDriverComms = NULL;
			fwprintf( LogFile, L"Unable to open communication channel to filter driver. Possibly see NET HELPMSG %d\n", GetLastError());
			wprintf( L"Unable to open communication channel to filter driver. Possibly see NET HELPMSG %d\n", GetLastError());
		} 
	}
	else
	{
		fwprintf( LogFile, L"Attempt to Load Registry Kernal Driver Failed. Possibly see NET HELPMSG %d\n", GetLastError());
		wprintf( L"Attempt to Load Registry Kernal Driver Failed. Possibly see NET HELPMSG %d\n", GetLastError());
	}

	return bRegDriverLoaded;
}



//------------------------------------------------------
//
//


bool RegEventsMonitor::InstallRegistryKernelDriverAsService()
{
	SC_HANDLE hSCManager;
	wchar_t DriverFullPathname[1024];
	int ErrorId = 0;
	
	GetFullPathName( REG_EVENT_FILTER_DRIVER_FILENAME, 1024, DriverFullPathname, NULL);


#ifdef _DEBUG

	//------------------------------------------------------
	// debug code here.. for debugging on a local VM.
	// this is to ensure we're not trying to load the filter driver
	// from the debug location, ie. the Z drive  ( share of the D drive
	// in the dev environment ). This may not
	// be accessible to th service process, and so cannot be
	// a service.

	if ( DriverFullPathname[0] == (wchar_t) L'z' || DriverFullPathname[0] == (wchar_t) L'Z')
	{
		wcscpy_s( DriverFullPathname, L"c:\\installmonitor\\CaptureRegistryMonitor.sys" );
	}

#endif


	//------------------------------------------------------
	// open the services control manager asking permission
	// to create a new service.

    hSCManager = OpenSCManager
				 (
					NULL,						// NULL = local machine.
					NULL,						// NULL = services active database
					SC_MANAGER_CREATE_SERVICE	// the access rights we want.
				 );


    if( hSCManager == NULL )
    {
		ErrorId =  GetLastError();
		fwprintf( LogFile, L"Unable to open the Service Control Manager \nPossibly see NET HELPMSG %d\n", ErrorId );
		wprintf( L"Unable to open the Service Control Manager \nPossibly see NET HELPMSG %d\n", ErrorId );
		return false;
	}


	//------------------------------------------------------
	// attempt to create the service.

    hService =	CreateService
				(
					hSCManager, 
					REG_EVENT_FILTER_DRIVER_NAME, 
					REG_EVENT_FILTER_DRIVER_NAME, 
                    SERVICE_START | DELETE | SERVICE_STOP,  // desired access rights to the service.
                    SERVICE_KERNEL_DRIVER,
                    SERVICE_DEMAND_START,					// dont auto start this service.
                    SERVICE_ERROR_IGNORE, 
					DriverFullPathname,						// path to the .sys file
                    NULL, NULL, NULL, NULL, NULL
				);


    if( hService != NULL || GetLastError() == ERROR_SERVICE_EXISTS )
    {
		if (  GetLastError() == ERROR_SERVICE_EXISTS )
		{
			fwprintf( LogFile, L"Registry kernel driver already exists. Continuing");
		}

		hService = OpenService
					( 
						hSCManager, 
						REG_EVENT_FILTER_DRIVER_NAME, 
						SERVICE_START | DELETE | SERVICE_STOP  // desired access rights to the service.
					);
    }


	CloseServiceHandle(hSCManager);

	// if we couldnt open the service, then give an error message and give up.

    if( hService == NULL )
	{
		ErrorId =  GetLastError();
		fwprintf( LogFile, L"Error loading registry kernel driver\nPossibly see NET HELPMSG %d\n", ErrorId );
		wprintf( L"Error loading registry kernel driver\nPossibly see NET HELPMSG %d\n", ErrorId );
		return false;
	}

	fwprintf( LogFile, L"Loaded kernel registry filter driver\n");

	return true;
}



//------------------------------------------------------------------
//
// Start the service / driver.
//


bool RegEventsMonitor::StartRegistryKernelDriverService()
{
    if(  StartService( hService, 0, NULL ) == FALSE )
	{
		DWORD err = GetLastError();

		if( err == ERROR_SERVICE_ALREADY_RUNNING )
		{
			fwprintf( LogFile, L"Registry Filter Driver service already running\n");
		} 
		else 
		{
			fwprintf( LogFile, L"Error starting service for the registry kernel driver : Possibly see NET HELPMSG %d\n", err);
			wprintf( L"Error starting service for the registry kernel driver : Possibly see NET HELPMSG %d\n", err);
			return false;
		}
	}
				
	return true;
}



//-------------------------------------------------------
// 
// Uninstall the Driver.
//

void RegEventsMonitor::UninstallRegistryKernelDriver()
{
	if( hService != NULL )
	{
		SERVICE_STATUS ss;

		ControlService(hService, SERVICE_CONTROL_STOP, &ss);
		
		if ( !DeleteService(hService) )
		{
			fwprintf( LogFile, L"Error deleting service for the registry kernel driver : Possibly see NET HELPMSG %d\n", GetLastError() );
		}
		
		CloseServiceHandle(hService);
	}

	hService = NULL;
}


