//----------------------------------------------------------------------------
//
//		Console.cpp
//
//		Kernel / User Mode	:	User
//
//		
//		Responsible for 
//		- creating the log file.
//		- creating an output directory.
//		- starting the file and registry monitor threads
//		- initiating an application installer
//		- compressing the results into a zip file
//		- deleting the output directory
//
// File History
//
//	Date			Person		Comment
//
//  1 Feb 2011		M Lewis		First Written
//
//----------------------------------------------------------------------------
// consolecpp.cpp : Defines the entry point for the console application.
//

#include "FileEventsMonitor.h"
#include "RegEventsMonitor.h"
#include "stdafx.h"
#include "io.h"
#include "ConfigFile.h"
#include "GenEnvData.h"

bool GetDriverLoadPrivilege();
bool CompressOutputDirectory( FILE* pLog, wstring Key );
void DeleteOutputDirectory();
bool isDirectory( wstring filePath );
void StartKeyClickProcess( FILE* pLog );
void CloseKeyClick( FILE* pLog, int iSecondsToWaitForKeyClick );
void WriteProgressText( wchar_t* text );

wstring Encode( wstring ClearKey );



// make the config file a global resource.

CConfigFile ConfigFile;


// filename and directory definitions for this app.

#define OUTPUT_DIR					L"output"
#define OUTPUT_WRITTEN_FILES_DIR	L"output\\written"
#define OUTPUT_DELETED_FILES_DIR	L"output\\deleted"
#define OUTPUT_BEFORE_FILES_DIR		L"output\\before"
//#define OUTPUT_DELETED_FILES_DIR	L"d:\\output"

#define FILES_XML_FILENAME			L"output\\files.xml"
#define REG_XML_FILENAME			L"output\\registry.xml"
#define ENV_XML_FILENAME			L"output\\env.xml"

#define LOG_FILENAME				L"Log.txt"

#define PROGRESS_FILENAME			L"Progress.txt"

#define CONFIG_FILENAME				L"InstallMonitor.ini"

#define COMPRESSED_RESULTS			L"results.cap"

// application entry point.


int _tmain(int argc, _TCHAR* argv[])
{

	FileEventsMonitor		*FileMonThread;
	RegEventsMonitor		*RegMonThread;

	STARTUPINFOW			siStartupInfo;
    PROCESS_INFORMATION		piProcessInfo;

	WCHAR					FullPathToExe[1000];
	wstring					UserProvidedKey;

	int						SecondsToWaitForAppInstaller; 
	int						consoleconfig;
	int						capturemode;
	int						raisepriority;
	int						keyclick_launch;
	int						SecondsToWaitForKeyClick;

	FILE*					LogFile = NULL;

	int						ReturnCode = 0;

	//-----------------------------------------------------

	WriteProgressText( L"Initialising.." );

	//-----------------------------------------------------
	// Open a new log file, with shared access.
	// This allows other apps read only access.
	// Useful if we want to load it into Notepad whilst this 
	// app is still running.
	// c flag is specified so that log entries are not lost
	// in event of app crash.

	LogFile = _wfsopen( LOG_FILENAME, L"wc", _SH_DENYWR );

	if ( LogFile == 0 )
	{	
		printf("\nUnable to open the log file\n");
		WriteProgressText( L"Log Error. Aborted." );
		exit(1);
	}

	//-----------------------------------------------------

	HRESULT hr = CoInitialize(NULL);
    
	if (!SUCCEEDED(hr))
    {
		fprintf( LogFile, "\nUnable to initialise COM.  Quitting \n");
		WriteProgressText( L"Com Error. Aborted." );
		exit(1);
	}

	//-----------------------------------------------------
    // parse the user input to find out what installer 
    // process we have to run..
    //
    // for this code example.. I'm assuming that
    //
    // there will be only two args
    //
	//	1.	a key of some sort to be used to encrypt the zip file.
	//  2.  a path to the working directory for the installer.
	//	3.	a command line for the app installer we want to monitor.
	//
    // argv[0] will be the name of this app. 
	// ie. installmonitor
    // 
	// argv[1] will be the key code.
	//
	// argv[2] will be the working dir
	//
	// argv[3..n] will be the full path to the install app that
    // has to be monitored.    
	// now get the arguments if present.

	
	if ( argc < 4 )	
	{
		printf("\nParameter missing. \n\n 1) a key code, 2) a working folder path and 3) a command line for the Installer App to be monitored. Quitting \n");
		fprintf( LogFile, "\nParameter missing. \n\n 1) a key code, 2) a working folder path and 3) a command line for the Installer App to be monitored. Quitting \n");
		WriteProgressText( L"Args Error. Aborted." );
		exit(1);
	}

	// initialise the vars to hold the params.

	wcscpy_s( FullPathToExe, 1000, L"" );
	UserProvidedKey = L"";
	

	//--------------------------------------------
	// the key is the 1st param

	UserProvidedKey.assign( argv[1] );

	UserProvidedKey = Encode( UserProvidedKey );

	//--------------------------------------------
	// the working folder dir is the 2nd param

	wstring InstallerWorkingDirectory;

	InstallerWorkingDirectory.assign( argv[2] );

	//------------------------------------------------------------------------------
    // test the working dir.

	if ( isDirectory( InstallerWorkingDirectory ) == false )
	{
		fwprintf( LogFile, L"\n'%s' is not a valid working dir. Will use TEMP area.\n", InstallerWorkingDirectory.c_str() );
		size_t  RetSize;
		wchar_t TempDir[1000];

		errno_t status = _wgetenv_s( &RetSize, TempDir, 1000, L"TEMP" );

		if ( status == 0 && RetSize != 0 )
		{
			InstallerWorkingDirectory.assign( TempDir );				
		}
		else
		{
			InstallerWorkingDirectory = L"c:\\";
		}

		fwprintf( LogFile, L"'%s' will be used for the working dir\n", InstallerWorkingDirectory.c_str() );		

	}

	//--------------------------------------------
	// now take the next 5 or so parameters (if provided) and join them together to
	// for a command line for the isntaller app.

    if ( argc >= 4 )
    {
		wcscpy_s( FullPathToExe, 1000, argv[3] );
    }

	if ( argc >= 5 )
    {
		wcscat_s( FullPathToExe, 1000, L" " );
		wcscat_s( FullPathToExe, 1000, argv[4] );
	}

	if ( argc >= 6 )
    {
		wcscat_s( FullPathToExe, 1000, L" " );
		wcscat_s( FullPathToExe, 1000, argv[5] );
	}

	if ( argc >= 7 )
    {
		wcscat_s( FullPathToExe, 1000, L" " );
		wcscat_s( FullPathToExe, 1000, argv[6] );
	}

	if ( argc >= 8 )
    {
		wcscat_s( FullPathToExe, 1000, L" " );
		wcscat_s( FullPathToExe, 1000, argv[7] );
	}

	if ( argc >= 9 )
    {
		wcscat_s( FullPathToExe, 1000, L" " );
		wcscat_s( FullPathToExe, 1000, argv[8] );
	}

	//------------------------------------------------------------------------------

	if( GetDriverLoadPrivilege() == false )
	{	
		printf("\nPlease run with Admin rights\n");
		fprintf( LogFile, "\nPlease run with Admin rights\n");
		WriteProgressText( L"Admin Rights Error. Aborted." );
		exit(1);
	}

	//-----------------------------------------------------

	if ( ConfigFile.LoadFile( CONFIG_FILENAME ) == false )
	{
		printf("\nUnable to open ini file\n\n");
		fprintf( LogFile, "\nUnable to open ini file\n\n");
		WriteProgressText( L" Ini File Error. Aborted." );
		exit(1);
	}

	ConfigFile.GetIntValueFor( L"consoleconfig", consoleconfig );

	ConfigFile.GetIntValueFor( L"capturemode", capturemode );

	ConfigFile.GetIntValueFor( L"keyclick_launch", keyclick_launch );
	
	ConfigFile.GetIntValueInRange( L"SecondsToWaitForKeyClick", SecondsToWaitForKeyClick, 0, 600, 10 );


	//------------------------------------------------------------------------------

	ConfigFile.GetIntValueFor( L"raisepriority", raisepriority );

	if ( raisepriority & 1 )
	{
		SetPriorityClass ( GetCurrentProcess(), ABOVE_NORMAL_PRIORITY_CLASS );
	}

	//------------------------------------------------------------------------------

	CreateDirectory( OUTPUT_DIR, NULL );

	//------------------------------------------------------------------------------
	// Grab the win.ini file. we may need it to determine whats changed during install

	CreateDirectory( OUTPUT_BEFORE_FILES_DIR, NULL );

	TCHAR szSrcPath[MAX_PATH];

	if ( SHGetFolderPath(NULL,  CSIDL_WINDOWS, NULL, 0, szSrcPath) == S_OK )
	{
		wstring SrcFilePath = szSrcPath;
		wstring DestFilePathAndName = OUTPUT_BEFORE_FILES_DIR;

		SrcFilePath = SrcFilePath + L"\\win.ini";
		DestFilePathAndName = DestFilePathAndName + L"\\win.ini";

		if ( CopyFile( SrcFilePath.c_str(), DestFilePathAndName.c_str(), FALSE ) == FALSE )
		{
			printf( "\nUnable to grab a copy of win.ini\n\n");
			fprintf( LogFile, "\nUnable to grab a copy of win.ini\n\n");
		}
	}
	else
	{
		printf( "\nUnable to find windows dir whilst grabbing copy of win.ini\n\n");
		fprintf( LogFile, "\nUnable to find windows dir whilst grabbing copy of win.ini\n\n");
	}

	//------------------------------------------------------------------------------
	// get a value for SecondsToWaitForAppInstaller from the ini file.
	//
	// set a min and max range of 30 and 100000. If entry not found use
	// a default of 600
	// 

	ConfigFile.GetIntValueInRange( L"SecondsToWaitForAppInstaller", SecondsToWaitForAppInstaller, 30, 100000, 600 );

	//------------------------------------------------------------------------------
	// start the keypress app.

	if ( keyclick_launch )
	{
		StartKeyClickProcess( LogFile );
	}

	//------------------------------------------------------------------------------

    FileMonThread = new FileEventsMonitor( LogFile, FILES_XML_FILENAME );

	FileMonThread->SetDeletedFilesDir( TRUE, OUTPUT_DELETED_FILES_DIR );

	FileMonThread->StartMonitorThread();

	fflush( LogFile );

	//------------------------------------------------------------------------------

	RegMonThread = new RegEventsMonitor( LogFile, REG_XML_FILENAME );

	RegMonThread->StartMonitorThread();

	fflush( LogFile );

	//------------------------------------------------------------------------------
	// code to start the application installer process.

    wprintf( L"\nstarting installer process %s\n", FullPathToExe );
	fwprintf( LogFile, L"\nstarting installer process '%s' with working dir of '%s'\n", FullPathToExe, InstallerWorkingDirectory.c_str() );

	fflush( LogFile );
    fflush( stdout );


    memset(&siStartupInfo, 0, sizeof(siStartupInfo));
    memset(&piProcessInfo, 0, sizeof(piProcessInfo));

	siStartupInfo.cb = sizeof(siStartupInfo);


    if (    CreateProcessW
			( 
				NULL, 
				FullPathToExe, 
				0, 
				0, 
				FALSE,
                CREATE_DEFAULT_ERROR_MODE, 
				0, 
				InstallerWorkingDirectory.c_str(),
                &siStartupInfo, 
				&piProcessInfo
			) != FALSE)
    {
        // now to send the Process ID to the filter so that it can separate out 
        // the writes done by our process

		FileMonThread->SetFilterByThreadID( piProcessInfo.dwProcessId );

		RegMonThread->SetFilterByThreadID( piProcessInfo.dwProcessId );
		
		if ( capturemode & 1 )
		{
			FileMonThread->bCaptureAllEvents = true;
			RegMonThread->bCaptureAllEvents = true;
		}

        fwprintf( LogFile, L"\ninstaller process '%s' started with process id = %d\n\n", FullPathToExe, piProcessInfo.dwProcessId );

        wprintf( L"\ninstaller process '%s' started.\n\n", FullPathToExe );

		WriteProgressText( L"Installer Running." );

		fflush( LogFile );


		//------------------------------------------------------------------------------
		// hide the console now if required.

		if ( consoleconfig == 0 ) 
		{
			Sleep(500);
			ShowWindow( GetConsoleWindow(), SW_HIDE );
		}
		else
		if ( consoleconfig == 1 ) 
		{
			Sleep(500);											// added here to stop an odd interaction with child process
			ShowWindow( GetConsoleWindow(), SW_MINIMIZE );
		}
	

		//------------------------------------------------------------------------------
		// wait here until our process finishes..  
        // for safety add in a sensible timeout value.

        if ( WaitForSingleObject( piProcessInfo.hProcess, (SecondsToWaitForAppInstaller * 1000)) == WAIT_TIMEOUT )
		{
            fwprintf( LogFile, L"\nInstaller process has taken too long. Killing App Installer and Quitting\n\n" );
            wprintf( L"\nInstaller process has taken too long. Killing App Installer and Quitting\n\n" );

			WriteProgressText( L"Installer Timeout. Aborted." );

			fflush( LogFile );

			// kill off the app installer process.

			DWORD dwExitCode = 0;
			HANDLE hHandle = ::OpenProcess( PROCESS_TERMINATE, 0, piProcessInfo.dwProcessId );
			::GetExitCodeProcess (hHandle, &dwExitCode );
			::TerminateProcess( hHandle, dwExitCode );

			ReturnCode = 3;
		}
		else
		{
			WriteProgressText( L"Installer Finished" );

			DWORD ExitCode = 0;
			if ( GetExitCodeProcess( piProcessInfo.hProcess, &ExitCode ) != 0 )
			{
				fwprintf( LogFile, L"\ninstaller process has finished. Exit Code = 0x%08x\n\n", ExitCode );
			}
			else
			{
				fwprintf( LogFile, L"\ninstaller process has finished. Unable to get exit code.\n\n" );
			}
		}
	}
	else
	{
		wchar_t ErrMsg[1000];
		ErrMsg[0] = 0;
		DWORD ErrorId = GetLastError();

		FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM, 0, ErrorId, 0, ErrMsg, 1000, 0 );

		WriteProgressText( L"Installer failed to start. Aborted." );

		fwprintf( LogFile, L"\nInstaller process failed to start. \n\nError Code = 0x%08x : %s\n\n", ErrorId, ErrMsg );
		wprintf( L"\nInstaller process failed to start. \n\nError Code = 0x%08x : %s\n\n", ErrorId, ErrMsg );
		ReturnCode = 2;
	}


	fflush( LogFile );

	//------------------------------------------------------------
	// now start the process of closing the monitoring stuff down.
	//
	// first wait a couple of seconds to let the buffers drain.
	//

	Sleep( 1000 );
	
	//------------------------------------------------------------
	// shutdown the keyclick process.

	if ( keyclick_launch )
	{
		CloseKeyClick( LogFile, SecondsToWaitForKeyClick );
	}

	//------------------------------------------------------------
	// shut down the file and registry monitoring threads.

	Sleep( 1000 );

	fwprintf( LogFile, L"\nUnloading registry kernal filter driver\n" );


	RegMonThread->StopMonitorThread();

	delete RegMonThread;

	//--------------------------------

	fwprintf( LogFile, L"Unloading file kernal filter driver\n\n" );

	FileMonThread->SetDeletedFilesDir( FALSE, L"" );

	FileMonThread->StopMonitorThread();

	fwprintf( LogFile, L"Copying modifed / created files.\n\n" );

	wprintf( L"Copying modifed / created files.\n\n" );

	fflush( LogFile );

	WriteProgressText( L"Copying Files" );

	FileMonThread->CopyModifiedFilesToOutputDir( OUTPUT_WRITTEN_FILES_DIR );

	if ( capturemode & 2 )
	{
		FileMonThread->DumpProcessRelationshipsCache();
	}

	delete FileMonThread;

	//------------------------------------------------------------

	GenEnvData EnvInfoGenerator;

	EnvInfoGenerator.GenerateEnvDataXML( LogFile, ENV_XML_FILENAME );

	//------------------------------------------------------------

	WriteProgressText( L"Compressing Results" );

	wprintf( L"Compressing output directory.\n\n" );

	fwprintf( LogFile, L"\nCompressing output directory.\n\n" );


	// important to flush the output to the file at this point.
	// if we dont, then the output from the compression subprocess 
	// will appear in the wrong place in the file.

	fflush( LogFile );

	if ( ! CompressOutputDirectory( LogFile, UserProvidedKey ) )
	{
		ReturnCode = 4;
	}

	DeleteOutputDirectory();

	//------------------------------------------------------------

	wprintf( L"end.\n" );

	fwprintf( LogFile, L"\nend.\n" );
	fclose( LogFile );

	WriteProgressText( L"Finished" );

	//------------------------------------------------------------

	CoUninitialize();

	return ReturnCode;
}




//------------------------------------------------------------
// routine to startup the KeyClick application
//------------------------------------------------------------

    PROCESS_INFORMATION		piKeyPressProcessInfo;
	BOOL					bKeyClickStarted;


void StartKeyClickProcess( FILE* pLog )
{
	WCHAR					KeyClickExe[1000];
	STARTUPINFOW			siKeyPressStartupInfo;

	bKeyClickStarted = FALSE;

	memset(&siKeyPressStartupInfo, 0, sizeof(siKeyPressStartupInfo));
	memset(&piKeyPressProcessInfo, 0, sizeof(piKeyPressProcessInfo));

	siKeyPressStartupInfo.cb = sizeof(siKeyPressStartupInfo);

	siKeyPressStartupInfo.lpTitle = L"AutoKeyClicker-ConvertIT";


	wcscpy_s( KeyClickExe, 1000, L"KeyClick.exe" );

	bKeyClickStarted =
			CreateProcessW
			( 
				NULL, 
				KeyClickExe, 
				0, 
				0, 
				FALSE,
				CREATE_NEW_CONSOLE, 
				0, 
				NULL,
				&siKeyPressStartupInfo, 
				&piKeyPressProcessInfo
			); 

	if ( bKeyClickStarted )
	{	
		fwprintf( pLog, L"KeyClick process started.\n\n" );
		wprintf( L"KeyClick process started.\n\n" );
	}
	else
	{	
		wchar_t ErrMsg[1000];
		ErrMsg[0] = 0;
		DWORD ErrorId = GetLastError();

		FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM, 0, ErrorId, 0, ErrMsg, 1000, 0 );

		fwprintf( pLog, L"\nKeyclick process failed to start. \n\nError Code = 0x%08x : %s\n\n", ErrorId, ErrMsg );
		wprintf( L"\nKeyclick process failed to start. \n\nError Code = 0x%08x : %s\n\n", ErrorId, ErrMsg );
	}
}



//------------------------------------------------------------
// Shutdown the KeyClick app gracefully
//------------------------------------------------------------


void CloseKeyClick( FILE* pLog, int iSecondsToWaitForKeyClick )
{
	if ( bKeyClickStarted == FALSE ) return;

	HWND hwnd = FindWindow( NULL, L"AutoKeyClicker-ConvertIT" );

	if ( hwnd == NULL )
	{
		fwprintf( pLog, L"Could not find window of keyclick to close it\n\n" );
		wprintf( L"Could not find window of keyclick to close it\n\n" );
	}
	else
	{
		fwprintf( pLog, L"Closing keyclick program\n\n" );
		wprintf( L"Closing keyclick program\n\n" );

		PostMessage( hwnd, WM_CHAR, 'q', 0 );
	}

	//------------------------------------------------------------
	// wait for the keyclick to terminate.

	if ( WaitForSingleObject( piKeyPressProcessInfo.hProcess, (iSecondsToWaitForKeyClick * 1000)) == WAIT_TIMEOUT )
	{
        fwprintf( pLog, L"\nKeyClick process has taken too long to close. Killing KeyClick\n\n" );
        wprintf( L"\nKeyClick process has taken too long. Killing KeyClick\n\n" );

		fflush( pLog );

		// kill off the app installer process.

		DWORD dwExitCode = 0;
		HANDLE hHandle = ::OpenProcess( PROCESS_TERMINATE, 0, piKeyPressProcessInfo.dwProcessId );
		::GetExitCodeProcess (hHandle, &dwExitCode );
		::TerminateProcess( hHandle, dwExitCode );
	}
	else
	{
		fwprintf( pLog, L"Closed Keyclick\n" );
		wprintf( L"Closed Keyclick\n" );
	}

}


//------------------------------------------------------------
//

void WriteProgressText( wchar_t* text )
{
	FILE*	pFile = NULL;

	pFile = _wfsopen( PROGRESS_FILENAME, L"w", _SH_DENYWR );

	if ( pFile == 0 ) return;

	fwprintf( pFile, L"%s", text );

	fclose( pFile );
}

//------------------------------------------------------------
//
// This application requires a special privilege to allow it
// to load drivers into the kernel.
//


bool GetDriverLoadPrivilege()
{
	// First get the access token handle

	HANDLE hAccessToken;

	BOOL bOK = OpenProcessToken
				( 
					GetCurrentProcess(),					// ProcessHandle
					TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY,	// DesiredAccess
					&hAccessToken							// A pointer to a handle that identifies the newly opened access token when the function returns.
				);
	

	if ( bOK == FALSE )
	{
		printf("Unable to open the process token. Error code = %u\n", GetLastError() ); 
		return false;
	}

	
	// Use LookupPrivilegeValue function to retrieve the locally unique 
	// identifier (LUID) used on a specified system to locally represent 
	// the specified privilege name.


	LUID				luid;
	LPCTSTR				lpszPrivilege = L"SeLoadDriverPrivilege";


	bOK = LookupPrivilegeValue
			( 
				NULL,           // lookup privilege on local system
				lpszPrivilege,  // privilege to lookup 
				&luid			// receives LUID of privilege
				);


	if ( bOK == FALSE )
	{
		printf("LookupPrivilegeValue error: %u\n", GetLastError() ); 
		return false; 
	}


	// Enable the privilege 

	TOKEN_PRIVILEGES			tp;

	tp.PrivilegeCount			= 1;
	tp.Privileges[0].Luid		= luid;
	tp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;

		
	bOK = AdjustTokenPrivileges
		  (
			hAccessToken,					// TokenHandle
			FALSE,							// DisableAllPrivileges
			&tp,							// NewState
			sizeof(TOKEN_PRIVILEGES),		// BufferLength 
			(PTOKEN_PRIVILEGES) NULL,		// PreviousState
			(PDWORD) NULL					// ReturnLength
		  );


	if ( bOK == FALSE )
	{ 
		printf("Error whilst calling AdjustTokenPrivileges : %u\n", GetLastError() );
		return false; 
	} 

	if (GetLastError() == ERROR_NOT_ALL_ASSIGNED)
	{	
		printf("Could not gain the required privilege. \n");
		return false;
	} 
	
	CloseHandle( hAccessToken );
	
	return true;
}



//------------------------------------------------------------
// Compresses all of the files in the 'output' directory
// and writes them into a zip file.


bool CompressOutputDirectory( FILE* pLog, wstring Key )
{
	BOOL				ProcCreated = FALSE;

	STARTUPINFO			siStartupInfo;
	PROCESS_INFORMATION piProcessInfo;

	// ensure that any previous results file is removed.

	DeleteFile( COMPRESSED_RESULTS );

	memset(&siStartupInfo, 0, sizeof(siStartupInfo));
	memset(&piProcessInfo, 0, sizeof(piProcessInfo));
	
	// put some entries into the StartupInfo structure to tell the createed process
	// to send its output text to the Log instead of to the screen.
	// This way we get to log the progress / and or errors from the zip operating.

	siStartupInfo.cb = sizeof(siStartupInfo);

	siStartupInfo.dwFlags    = STARTF_USESTDHANDLES;

	siStartupInfo.hStdError  = (HANDLE) _get_osfhandle(_fileno(pLog));   
	siStartupInfo.hStdOutput = (HANDLE) _get_osfhandle(_fileno(pLog));   

	// create the commmand line required.
	// we will be setting the working directory of the process to be within
	// the output sub directory itself.
	// hence we ask that the resulting zip file is placed in the parent
	// dir. ie. the main apppliction dir.

	wchar_t cmd[100];
	
//	wsprintf( cmd, L"\"7za.exe\" a -tzip -y -p%s \"..\\%s\" * ", Key.c_str(), COMPRESSED_RESULTS ); 

	//	wsprintf( cmd, L"\"7za.exe\" a -tzip -y  \"..\\%s\" * ", COMPRESSED_RESULTS ); 

	HRESULT hr = StringCbPrintf( cmd, 100*sizeof(wchar_t), L"\"7za.exe\" a -tzip -y -p%s \"..\\%s\" * ", Key.c_str(), COMPRESSED_RESULTS ); 

//	HRESULT hr = StringCbPrintf( cmd, 100*sizeof(wchar_t), L"\"7za.exe\" a -tzip -y  \"..\\%s\" * ", COMPRESSED_RESULTS ); 

	if ( hr != S_OK )
	{
		fwprintf( pLog, L"Cannot create the comand line for compression process : 0x%08x\n", GetLastError());
		return false;
	}

	// set the 'working dir' to be the 'output' sub dir.

	wchar_t WorkingDirectory[1024];
	
	GetFullPathName( OUTPUT_DIR, 1024, WorkingDirectory, NULL);


	ProcCreated = CreateProcess( NULL,
								(LPTSTR) cmd, 
								0, 
								0, 
								TRUE,
								CREATE_DEFAULT_ERROR_MODE, 
								0, 
								WorkingDirectory, 
								&siStartupInfo,
								&piProcessInfo
								);


	if( ProcCreated == FALSE )
	{
		fwprintf( pLog, L"Cannot create the compression process - error : 0x%08x\n", GetLastError());
		return false;
	} 


	WaitForSingleObject( piProcessInfo.hProcess, INFINITE );

	return true;
}



//------------------------------------------------------------



void DeleteOutputDirectory()
{
		wchar_t			szFullPath[1024];

		SHFILEOPSTRUCT	FileOperation;

		GetFullPathName( OUTPUT_DIR, 1024, szFullPath, NULL);

		FileOperation.wFunc					= FO_DELETE;
		FileOperation.hwnd					= NULL;
		FileOperation.pTo					= NULL;
		FileOperation.fAnyOperationsAborted	= FALSE;
		FileOperation.lpszProgressTitle		= NULL;
		FileOperation.hNameMappings			= NULL;
		FileOperation.pFrom					= szFullPath;
		FileOperation.fFlags				= FOF_NOCONFIRMATION | FOF_SILENT | FOF_NOERRORUI;

		SHFileOperation( &FileOperation );
}


//---------------------------------------------------------
// tests to see if a path refers to a directory.
//


bool isDirectory( wstring filePath )
{
	DWORD code = GetFileAttributes(filePath.c_str());

	if ( code == INVALID_FILE_ATTRIBUTES ) return false;

	if ( ( code & FILE_ATTRIBUTE_DIRECTORY ) == 0 ) return false;
	
	return true;
}


//----------------------------------------------------------------
// Code to take the key passed in as 1st arg to this app,
// and then mash it up to form an encryption key for the zip library
//
// Based on the following code by Sean.

/*
An example:
        Dim v As String =  "EXEPUTESKNARFBOBSEGAK"
        Dim enc As String = aokConvertItPlus.KeyMorph.Encode(v)  Returns "OHJFUIQMJYEUNTG[W[I_I" and this is your zip password
        Dim dec As String = aokConvertItPlus.KeyMorph.Decode(enc)  Returns "EXEPUTESKNARFBOBSEGAK"


Public Class KeyMorph

    Private Shared mKey As Integer() = New Integer() {4, 7, 3, 1, 2, 7, 2, 11, 4, 7}


    Public Shared Function Encode(ByVal Value As String) As String
        Dim Newstr As String = String.Empty
        Dim lenKey As Integer = mKey.Length
        Dim KeyPos As Integer = 0

        Value = New String(Value.Reverse().ToArray())
        For x As Integer = 0 To Value.Length - 1
            Newstr = Newstr + Chr(Asc(Value(x)) + mKey(KeyPos))
            KeyPos = KeyPos + 1
            If KeyPos >= lenKey Then KeyPos = 0
        Next
        Return Newstr
    End Function

    Public Shared Function Decode(ByVal Value As String) As String

        Dim Newstr As String = String.Empty
        Dim lenKey As Integer = mKey.Length
        Dim KeyPos As Integer = 0

        Value = New String(Value.Reverse().ToArray())
        For x As Integer = Value.Length - 1 To 0 Step -1
            Newstr = Newstr + Chr(Asc(Value(x)) - mKey(KeyPos))
            KeyPos = KeyPos + 1
            If KeyPos >= lenKey Then KeyPos = 0
        Next
        Return New String(Newstr.Reverse().ToArray())

    End Function

End Class

*/



wstring Encode( wstring ClearKey )
{
	int mKey[] =  {4, 7, 3, 1, 2, 7, 2, 11, 4, 7};

	wstring NewStr = L"";
	
	int lenkey = sizeof( mKey ) / sizeof( mKey[0] );

	int KeyPos = 0;

	wstring Value = ClearKey;
	reverse(Value.begin(), Value.end());		// reverse the string.

	for( int i = 0; i < Value.length(); i++ )
	{
		wchar_t c = Value.at( i );
		c = c + mKey[KeyPos];
		NewStr = NewStr + c;
		KeyPos++;
		if ( KeyPos >= lenkey ) KeyPos = 0;
	}

	return NewStr;
}


//----------------------------------------------------------------
